from typing import Dict, List, Any, Callable, Set

import os
import logging
import argparse
import asyncio
import shutil
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor

import datetime
import json
import functools
from dataclasses import dataclass

from sqlalchemy import or_

import zmq
from zmq.asyncio import Context as ZmqContext
from zmq.asyncio import Socket as ZmqSocket

import rx
import rx.subject
from rx import operators as ops
from rx.scheduler.eventloop import AsyncIOScheduler

from package.shared.constants import Constants, StreamSetting
from package.shared import models

from package.shared.logger import setup_logger
from package.shared.config import Config
from package.shared.ipc import IPCListener
from package.shared.flow_control import FlowControl
from package.shared.handlers import ESDataHandler
from package.shared.util import board_cfg_file

from package.shared.bridge import run_bridge
from package.board.vid_mgnt import run_vid_mgnt
from package.board.web import run_web
from package.board.mjpeg import run_mjpeg
from package.board.ad_server import run_adserver
from package.board.pos_client import run_posclient
from package.board.sync import run_sync


class ClientInfo():
    def __init__(self, client_id: bytes):
        self.m_client_id = client_id
        self.m_time = datetime.datetime.now()

    @property
    def client_id(self) -> bytes:
        return self.m_client_id

    @property
    def time(self) -> datetime.datetime:
        return self.m_time


class CleanHandler():
    async def handler(self, lifetime) -> None:
        logging.info('Start cleaning(lifetime: {})'.format(lifetime))

        now: datetime.datetime = datetime.datetime.now()
        clean_point: datetime.datetime = now - datetime.timedelta(seconds=lifetime)

        session = models.SessionSingleton.instance().session

        try:
            chunks: List[models.Chunk] = session.query(models.Chunk) \
                .filter(models.Chunk.created_at<clean_point) \
                .all()

            for chunk in chunks:
                chunk: models.Chunk
                filename: str = chunk.filename
                logging.info('Removing chunk: {}'.format(filename))

                try:
                    os.remove(filename)
                except:
                    logging.exception(f"Removing file error: {filename}")

            session.query(models.Chunk).filter(models.Chunk.created_at<clean_point).delete()
            session.query(models.Pos).filter(models.Pos.created_at<clean_point).delete()

            session.commit()
        except:
            logging.exception("Clean expired chunks failed")
            session.rollback()

    def on_next(self, obj):
        lifetime = obj['lifetime']
        asyncio.create_task(self.handler(lifetime))

    def on_error(self):
        pass

    def on_completed(self):
        pass


class CleanDiskusageHandler():
    async def handler(self, num_chunks) -> None:
        session = models.SessionSingleton.instance().session

        try:
            chunks: List[models.Chunk] = session.query(models.Chunk) \
                .order_by(models.Chunk.created_at.desc()) \
                .limit(num_chunks) \
                .all()

            for chunk in chunks:
                chunk: models.Chunk
                filename: str = chunk.filename
                logging.info('Removing chunk: {}'.format(filename))

                try:
                    os.remove(filename)
                except:
                    logging.exception(f"Removing file error: {filename}")

            session.query(models.Chunk) \
                .order_by(models.Chunk.created_at.desc()) \
                .limit(num_chunks) \
                .delete()

            session.commit()
        except:
            logging.exception("Clean disk usage failed")
            session.rollback()

    def on_next(self, obj):
        num_chunks = obj['num_chunks']
        asyncio.create_task(self.handler(num_chunks))

    def on_error(self):
        pass

    def on_completed(self):
        pass

class ServerToClientForwardHandler():
    def __init__(self, queue, router):
        self.m_queue = queue
        self.m_router = router

    async def handle(self, obj):
        msg_id = obj['id']
        if msg_id not in self.m_queue:
            logging.warning('Message ID: {} not in queue!'.format(msg_id))
            return

        client_id = self.m_queue[msg_id].client_id

        try:
            buf = json.dumps(obj).encode()
        except:
            logging.exception('Object to json conversion error')
            return

        logging.info('Forwarding {} ({}) from server to {}'.format(msg_id, obj['cmd'], client_id))
        await self.m_router.send_multipart([ client_id, buf ])

    def on_next(self, pkt):
        asyncio.create_task(self.handle(pkt['obj']))

    def on_error(self):
        pass

    def on_completed(self):
        pass


@dataclass
class DataChangedHandler():
    rxevents: rx.subject.Subject = None
    router: ZmqSocket = None

    async def handle(self, obj):
        if (self.rxevents is None) or (self.router is None):
            return

        clients = [ 'pos_client', 'vid_mgnt' ]

        msg = { 'cmd': 'notify-datachanged' }
        try:
            buf = json.dumps(msg).encode()
        except:
            logging.exception('Cant dumps obj to json')
            return

        for client in clients:
            await self.router.send_multipart([ client.encode(), buf ])

    def on_next(self, obj):
        asyncio.create_task(self.handle(obj))

    def on_error(self):
        pass

    def on_completed(self):
        pass


@dataclass
class Communicator(IPCListener):
    m_ctx: ZmqContext = None
    m_name: str = None
    m_queue: Dict[str, ClientInfo] = None
    m_clients: Set = None
    m_scheduler: AsyncIOScheduler = None
    m_rxevents: rx.subject.Subject = None

    args: Dict[str, str] = None
    executor: ProcessPoolExecutor = None


    def __post_init__(self):
        super().__post_init__()


    def setup(self):
        assert self.args is not None

        ctx = mp.get_context('spawn')
        self.executor = ProcessPoolExecutor(
            max_workers=40,
            mp_context=ctx)

        # Zmq
        self.setup_sockets()
        self.setup_ipc(url=self.args['ipc'], ctx=self.m_ctx, type=zmq.ROUTER)

        self.m_name = self.args['name']
        self.m_queue = dict()
        self.m_clients = set()

        # RX
        self.m_scheduler = AsyncIOScheduler(asyncio.get_running_loop())
        self.m_rxevents = rx.subject.Subject()

        self.m_rxevents.pipe(
            ops.filter(lambda obj: obj['cmd'] == 'clean')
        ).subscribe(CleanHandler(), scheduler=self.m_scheduler)

        self.m_rxevents.pipe(
            ops.filter(lambda obj: obj['cmd'] == 'clean_disk')
        ).subscribe(CleanDiskusageHandler(), scheduler=self.m_scheduler)

        self.m_rxevents.pipe(
            ops.filter(lambda obj: obj['cmd'] == 'forward_to_client')
        ).subscribe(ServerToClientForwardHandler(self.m_queue, self.ipc_sock), scheduler=self.m_scheduler)

        self.m_rxevents.pipe(
            ops.filter(lambda obj: obj['cmd'] == 'es-data')
        ).subscribe(ESDataHandler(self.m_rxevents), scheduler=self.m_scheduler)

        self.m_rxevents.pipe(
            ops.filter(lambda obj: obj['cmd'] == 'notify-datachanged')
        ).subscribe(DataChangedHandler(rxevents=self.m_rxevents, router=self.ipc_sock), scheduler=self.m_scheduler)


    def cleanup(self):
        self.close_ipc()
        self.clean_sockets()
        self.executor.shutdown(wait=False)


    async def handle_server_msg(self, buf: bytes) -> None:
        try:
            obj: Dict[str, Any] = json.loads(buf)
        except Exception as e:
            logging.warning('JSON Exception ({}): {}'.format(type(e), e))
            return

        if 'id' in obj:
            pkt = {
                'cmd': 'forward_to_client',
                'obj': obj
            }

            self.m_rxevents.on_next(pkt)
        else:
            logging.warning('Unknow to handle message from server:')
            logging.warning(obj)


    async def handle_ipc_async(self, client, buf):
        logging.info('Recv IPC msg from: {}'.format(client))

        try:
            obj: Dict[str, Any] = json.loads(buf)
        except Exception as e:
            logging.warning('JSON Exception ({}): {}'.format(type(e), e))
            return

        if obj['cmd'] == 'es-data':
            self.m_rxevents.on_next(obj)
            return
        elif obj['cmd'] == 'notify-datachanged':
            self.m_rxevents.on_next(obj)
            return
        else:
            logging.warning(f'Unhandled IPC message from client: {obj}')


    def handle_ipc(self, client, buf) -> None:
        asyncio.create_task(self.handle_ipc_async(client, buf))


    async def cleaner(self) -> None:
        config = Config.get()
        location = config.record.location

        while True:
            total, used, free = shutil.disk_usage(location)
            if free < config.record.diskspace_limit:
                logging.warning(f"Running out of space, remain {free}bytes, cleaning up")
                self.m_rxevents.on_next({ 'cmd': 'clean_disk', 'num_chunks': 500 })
            else:
                self.m_rxevents.on_next({ 'cmd': 'clean', 'lifetime': self.args['lifetime'] })
            await asyncio.sleep(int(self.args['cleaner']))


    async def run_process(self, name, func, **kwargs):
        if kwargs:
            func = functools.partial(func, **kwargs)

        loop = asyncio.get_running_loop()
        assert loop is not None

        try:
            await loop.run_in_executor(self.executor, func)
        except:
            logging.exception(f'Run process "{name}" exception')


    def setup_sockets(self) -> bool:
        self.m_ctx = ZmqContext()
        return True


    def clean_sockets(self) -> None:
        if self.m_ctx:
            self.m_ctx.term()
            self.m_ctx = None


    async def run_async(self) -> None:
        self.setup()

        # Set zmq_port to None
        session = models.SessionSingleton.instance().session
        try:
            session.query(models.Stream).update({ models.Stream.zmq_port: None })
            session.commit()
        except:
            logging.exception("Failed to clear zmq_port")
            session.rollback()

        funcs = [
            run_bridge,
            run_web,
            run_vid_mgnt,
            run_mjpeg,
            run_adserver,
            run_posclient
        ]

        tasks = [ self.run_process(str(f), f) for f in funcs ]

        try:
            await asyncio.gather(
                self.run_ipc(),
                self.cleaner(),
                run_sync(),
                *tasks)
        except:
            logging.exception("Board's tasks exception")

        self.cleanup()


def parse_argv():
    config = Config.get()

    parser = argparse.ArgumentParser()
    parser.add_argument('--name', help='Name of this communicator', default=config.app.gateway)

    parser.add_argument('--pub', help='Pub interface', default=config.bridge.internal)
    parser.add_argument('--ipc', help='IPC location', default=config.app.ipc)

    # Cleaner
    parser.add_argument('--lifetime', help='Remove chunks older this period of time (seconds)', default=config.record.lifetime)
    parser.add_argument('--cleaner', help='Run cleaner in every (seconds)', default=config.record.cleaner)

    # Clean db
    parser.add_argument('--cleandb', help='Cleanup database', default='false')

    return parser.parse_args()


async def run(args):
    args_dict = vars(args)

    app = Communicator(args=args_dict)
    await app.run_async()


def main(args):
    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    try:
        asyncio.run(run(args))
    except KeyboardInterrupt:
        logging.info('Done')


def cleandb():
    print('Cleaning up database')
    try:
        asyncio.run(models.cleanup())
    except Exception as e:
        print(e)


if __name__ == '__main__':
    Config.init(board_cfg_file())
    setup_logger('board.txt')

    args = parse_argv()
    if args.cleandb == 'true':
        cleandb()
    else:
        main(args)