import struct
import logging

from collections import namedtuple
from enum import Enum
from dataclasses import dataclass
from typing import Dict, List, Callable

class Type(Enum):
    TVT_LOGIN           = int('00000101', 16)
    TVT_LOGIN_OK        = int('00010001', 16)
    TVT_LOGIN_FAILED    = int('00010002', 16)
    TVT_VIDEO_START     = int('00000201', 16)
    TVT_VIDEO_FRAME     = int('0a000001', 16) # Additional header
    TVT_UNK1            = int('09000004', 16)
    TVT_SEMAPHORE       = int('09000003', 16)
    TVT_MOTION          = int('09000002', 16)
    TVT_VLOSS           = int('09000001', 16)
    TVT_CHANNEL_CHANGE  = int('00000202', 16) # Additional header
    TVT_PACK            = int('4b434150', 16)
    TVT_DVR_INFO        = int('09000008', 16)

    TVT_HEAD            = int('31313131', 16)
    TVT_HELLO_HEAD      = int('64616568', 16)
    TVT_I_FRAME	        = int('00000000', 16)
    TVT_P_FRAME	        = int('00000001', 16)

# Structs
# I: Head
# I: Size
STRUCT_header               = '<II'
SIZE_header                 = struct.calcsize(STRUCT_header)
Header                      = namedtuple('Header', [ 'type', 'size' ])
assert SIZE_header==8

# I   : Head
# 60s : unknow
STRUCT_hello_header         = '<I60s'
SIZE_hello_header           = struct.calcsize(STRUCT_hello_header)
HelloHeader                 = namedtuple('HelloHeader', [ 'type', 'unk1' ])
assert SIZE_hello_header==64

# I: Type
# I: Unknow
# I: mode
# I: Size
STRUCT_additional_header    = '<IIII'
SIZE_additional_header      = struct.calcsize(STRUCT_additional_header)
AdditionalHeader            = namedtuple('AdditionalHeader', [ 'type', 'unk1', 'mode', 'size' ])
assert SIZE_additional_header==16

# I  : Unknow
# I  : Unknow
# 36s: Username
# 36s: Password
# 28s: ClientId
# I  : Unknow
# I  : Unknow
# I  : Unknow
STRUCT_login_request        = '<II36s36s28sIII'
SIZE_login_request          = struct.calcsize(STRUCT_login_request)
LoginRequest                = namedtuple('LoginRequest', [ 'unk1', 'unk2', 'username', 'password', 'clientid', 'unk3', 'unk4', 'unk5' ])
assert SIZE_login_request==120

# I   : Unknow
# I   : Channel master
# I   : Unknow
# I   : Channel sub
# 20s : Unknow
STRUCT_start_request        = '<IIII20s'
SIZE_start_request          = struct.calcsize(STRUCT_start_request)
StartRequest                = namedtuple('StartRequest', [ 'unk1', 'master', 'unk2', 'sub', 'unk3' ])
assert SIZE_start_request==36

# I : Type
# I : Unknow
# I : Size
# I : Width
# I : Height
# I : Unknow
# I : Channel
# I : Unknow
# I : Unknow
# I : Unknow
# I : Unknow
# I : Unknow
# I : Unknow
# I : Unknow
# I : Unknow
STRUCT_frame_header         = '<IIIIIIIIIIIIIII'
SIZE_frame_header           = struct.calcsize(STRUCT_frame_header)
FrameHeader                 = namedtuple('FrameHeader', [ 'type', 'unk1', 'size', 'width', 'height', 'unk2', 'channel', 'unk3', 'unk4', 'unk5', 'unk6', 'unk7', 'unk8', 'unk9', 'unk10' ])
assert SIZE_frame_header==60

# I : Unknow
# I : Channel mask
# I : Unknow
STRUCT_event                = '<III'
SIZE_event                  = struct.calcsize(STRUCT_event)
Event                       = namedtuple('Event', [ 'unk1', 'channel_mask', 'unk2' ])
assert SIZE_event==12

# I : Number
# I : Size
# I : Unknow
STRUCT_pack                 = '<III'
SIZE_pack                   = struct.calcsize(STRUCT_pack)
Pack                        = namedtuple('Pack', [ 'num', 'size', 'unk2' ])
assert SIZE_pack==12


def type_of_buf(buf):
    btype, = struct.unpack('<I', buf[0:4])

    try:
        btype = Type(btype)
        return btype
    except:
        raise ValueError(f'{hex(btype)} is not a valid Type')


def read_header(buf):
    return Header(*struct.unpack(STRUCT_header, buf[0:SIZE_header]))


def create_header(header: Header):
    return struct.pack(STRUCT_header, header.type, header.size)


def read_additional_header(buf):
    return AdditionalHeader(*struct.unpack(STRUCT_additional_header, buf[0:SIZE_additional_header]))


def create_additional_header(adheader: AdditionalHeader):
    return struct.pack(STRUCT_additional_header, adheader.type, adheader.unk1, adheader.mode, adheader.size)


def create_start_request(start: StartRequest):
    return struct.pack(STRUCT_start_request, start.unk1, start.master, start.unk2, start.sub, start.unk3)

def read_pack(buf):
    return Pack(*struct.unpack(STRUCT_pack, buf[:SIZE_pack]))


def read_frame_header(buf):
    return FrameHeader(*struct.unpack(STRUCT_frame_header, buf[0:SIZE_frame_header]))


def create_login_request(req: LoginRequest):
    return struct.pack(STRUCT_login_request, req.unk1, req.unk2, req.username, req.password, req.clientid, req.unk3, req.unk4, req.unk5)


import asyncio

HandleOutput = namedtuple('HandleOutput', [ 'channel', 'frame', 'data', 'consumed' ])
HeaderOutput = namedtuple('HeaderOutput', [ 'size', 'data', 'additional' ])

# PACKET_NUM = 0

@dataclass
class TvtHandler:
    writer          : asyncio.StreamWriter
    username        : str
    password        : str
    channel         : int

    authenticated   : bool = False
    frame_count     : Dict[int, int] = None
    buffer          : memoryview = None


    def __post_init__(self):
        self.frame_count = dict()


    async def authenticate(self):
        logging.info('Authenticating')

        # head->head = TVT_HEAD;
        # head->size = sizeof(*login) + sizeof(*addhead);
        header: Header = Header(
            type=Type.TVT_HEAD.value,
            size=SIZE_login_request + SIZE_additional_header)

        # addhead->type = TVT_LOGIN;
        # addhead->unk1 = 0xffffffff;
        # addhead->unk2 = 0x00000004;
        # addhead->size = sizeof(*login);
        adheader: AdditionalHeader = AdditionalHeader(
            type=Type.TVT_LOGIN.value,
            unk1=int('ffffffff', 16),
            mode=int('00000004', 16),
            size=SIZE_login_request)

        # login->unk1 = 0x03;
        # login->unk2 = 0x00;
        # Q_ASSERT(username_.size() <= sizeof(login->user));
        # Q_ASSERT(password_.size() <= sizeof(login->password));
        # memcpy(login->user, username_.toLocal8Bit().constData(), username_.size());
        # memcpy(login->password, password_.toLocal8Bit().constData(), password_.size());
        # login->unk3 = 0xf4241b00;
        # login->unk4 = 0x00009ec0;
        # login->unk4 = 0x04;

        # 'unk1', 'unk2', 'username', 'password', 'clientid', 'unk3', 'unk4', 'unk5'
        login: LoginRequest = LoginRequest(
            unk1=int('03', 16),
            unk2=int('00', 16),
            username=self.username.encode(),
            password=self.password.encode(),
            clientid=b'',
            unk3=int('f4241b00', 16),
            unk4=int('00009ec0', 16),
            unk5=int('00000004', 16))

        buffer: bytearray = bytearray()
        buffer.extend(create_header(header))
        assert len(buffer)==SIZE_header

        buffer.extend(create_additional_header(adheader))
        assert len(buffer)==SIZE_header + SIZE_additional_header

        buffer.extend(create_login_request(login))
        assert len(buffer)==SIZE_header + SIZE_additional_header + SIZE_login_request

        self.writer.write(buffer)
        await self.writer.drain()


    async def request_channel(self):
        logging.info('Requesting channel')

        # head->head = TVT_HEAD;
        # head->size = sizeof(*start) + sizeof(*addhead);
        header: Header = Header(
            type=Type.TVT_HEAD.value,
            size=SIZE_additional_header + SIZE_start_request)

        # if (!haveSentChannelRequest_)
        #{
        #    addhead->type = TVT_VIDEO_START;
        #    haveSentChannelRequest_ = true;
        #}
        #else
        #    addhead->type = TVT_CHANNEL_CHANGE;

        # addhead->unk1 = 0x00;
        # addhead->unk2 = 0x00;
        # addhead->size = sizeof(*start);
        adheader: AdditionalHeader = AdditionalHeader(
            type=Type.TVT_VIDEO_START.value,
            unk1=0,
            mode=0,
            size=SIZE_start_request)

        channel_mask = 1 << (self.channel)

        # start->unk1 = 0x00;
        # start->unk2 = 0x0;;
        # start->channel_master = 0;
        # start->channel_sub = chanmask_;
        start: StartRequest = StartRequest(
            unk1=0,
            unk2=0,
            master=0,
            sub=channel_mask,
            unk3=b''
        )

        request: bytearray = bytearray()
        request.extend(create_header(header))
        assert len(request)==SIZE_header

        request.extend(create_additional_header(adheader))
        assert len(request)==SIZE_header+SIZE_additional_header

        request.extend(create_start_request(start))
        assert len(request)==SIZE_header+SIZE_additional_header+SIZE_start_request

        self.writer.write(request)
        await self.writer.drain()


    async def handle(self, buffer: memoryview):
        # consumed = 0
        offset = 0
        buf: memoryview = buffer

        def advance(size):
            nonlocal offset
            nonlocal buf

            offset += size
            buf = buf[size:]


        # Return:
        # -1  if error happend
        # 0   if everything ok (no mid chunk)
        # > 0 if an offset is needed (offset to mid chunk)
        def header_pos_in_data(start, size):
            nonlocal offset
            nonlocal buf

            if len(buf) < size:
                return -1

            j = -1
            for i in range(size, 0, -4):
                next_chunk = offset+i
                nbuf = buf[next_chunk:next_chunk+4]

                try:
                    ntype = type_of_buf(nbuf)
                    if ntype == Type.TVT_HEAD:
                        j = i
                        break
                except ValueError:
                    continue

            if j>=0:
                return j

            raise ValueError('Next chunk is not HEAD but not in current data neither')


        while True:
            # global PACKET_NUM
            # PACKET_NUM += 1

            packed = False
            size = 0

            if len(buf) < 4:
                break

            # Exception should be catched outside
            htype = type_of_buf(buf[size:size+4])

            # TVT_HELLO_HEAD
            if htype == Type.TVT_HELLO_HEAD:
                if len(buf) < size + SIZE_hello_header:
                    break

                await self.authenticate()
                yield HandleOutput(channel=None, frame=None, data=None, consumed=offset+SIZE_hello_header)
                return

            elif htype != Type.TVT_HEAD:
                raise ValueError(f'Expected TVT_HEAD but got {htype}')

            if len(buf) < size + SIZE_header:
                break

            assert htype == Type.TVT_HEAD

            # TVT_HEAD
            if len(buf) < size + SIZE_header:
                break

            header = read_header(buf[size:size+SIZE_header])
            size += SIZE_header
            data_size = 0

            if len(buf) < size+header.size:
                break

            if header.size > 0:
                adheader = read_additional_header(buf[size:size+SIZE_additional_header])
                size += SIZE_additional_header

                if adheader.type == Type.TVT_PACK.value:
                    assert len(buf) >= size+SIZE_pack

                    packed = True
                    pack = read_pack(buf[size:size+SIZE_pack])

                    size += SIZE_pack
                    assert pack.size != 0

                    if pack.num==1:
                        adheader = read_additional_header(buf[size:size+SIZE_additional_header])
                        size += SIZE_additional_header
                    else:
                        data_size = pack.size


                if not packed or pack.num==1:
                    if adheader.type == Type.TVT_VIDEO_FRAME.value:
                        if len(buf) < size + SIZE_frame_header:
                            break

                        fheader = read_frame_header(buf[size:size+SIZE_frame_header])
                        if fheader.type == Type.TVT_I_FRAME.value:
                            pass
                        elif fheader.type == Type.TVT_P_FRAME.value:
                            pass
                        else:
                            raise ValueError('Invalid frame type')

                        size += SIZE_frame_header
                        payload_size = size - SIZE_header
                        data_size = header.size - payload_size

                        assert len(buf) >= size + data_size
                    elif adheader.type == Type.TVT_MOTION.value:
                        logging.info('TVT_MOTION')
                        advance(SIZE_header + header.size)
                        continue
                    elif adheader.type == Type.TVT_VLOSS.value:
                        logging.info('TVT_VLOSS')
                        advance(SIZE_header + header.size)
                        continue
                    elif adheader.type == Type.TVT_UNK1.value:
                        logging.info('TVT_UNK1')
                        advance(SIZE_header + header.size)
                        continue
                    elif adheader.type == Type.TVT_LOGIN_OK.value:
                        logging.info('TVT_LOGIN_OK')
                        await self.request_channel()
                        size += adheader.size
                        yield HandleOutput(channel=None, frame=None, data=None, consumed=offset+size)
                        advance(size)
                        continue
                    elif adheader.type == Type.TVT_LOGIN_FAILED.value:
                        raise Exception('Invalid credentials')
                    elif adheader.type == Type.TVT_SEMAPHORE.value:
                        logging.info('TVT_SEMAPHORE')
                        size += adheader.size
                        yield HandleOutput(channel=None, frame=None, data=None, consumed=offset+size)
                        advance(size)
                        continue
                    elif adheader.type == Type.TVT_DVR_INFO.value:
                        logging.info('TVT_DVR_INFO')
                        size += adheader.size
                        yield HandleOutput(channel=None, frame=None, data=None, consumed=offset+size)
                        advance(size)
                        # PACKET_NUM = 0
                        continue
                    else:
                        raise ValueError(f'Unknow type: { hex(adheader.type) }')
            else:
                logging.warning('TVT: Size in header is zero')
                yield HandleOutput(channel=None, frame=None, data=None, consumed=offset+size)
                advance(size)
                continue

            data = None
            if data_size > 0:
                data = buf[size:size+data_size]

            yield HandleOutput(channel=self.channel, frame=None, data=data, consumed=offset+size+data_size)
            advance(size+data_size)

        # Wait for more data
        yield HandleOutput(channel=None, frame=None, data=None, consumed=offset)


@dataclass
class App():
    callback    : Callable
    host        : str
    port        : int
    channel     : int
    username    : str
    password    : str

    reader      : asyncio.StreamReader = None
    writer      : asyncio.StreamWriter = None
    read_size   : int = None
    buffer      : bytearray = None
    handler     : TvtHandler = None
    running     : bool = True
    skip        : int = 0
    total       : int = 0


    def __post_init__(self):
        self.buffer     = bytearray()
        self.read_size  = 512


    async def process_frame(self, op: HandleOutput):
        if self.callback and self.channel==op.channel:
            await self.callback(op)


    async def new_data_arrived(self):
        consumed: int = 0

        # Bytes to skip in buffer
        if self.skip>0:
            bsize = len(self.buffer)
            if bsize>self.skip:
                self.buffer = self.buffer[self.skip:]
                self.skip = 0
            else:
                self.skip -= bsize
                self.buffer = bytearray()
                return

        async for rs in self.handler.handle(memoryview(self.buffer)):
            consumed = rs.consumed

            if rs.channel is None:
                continue

            await self.process_frame(rs)

        if consumed>0:
            self.total += consumed

            bsize = len(self.buffer)
            if bsize < consumed:
                self.buffer = bytearray()
                self.skip = consumed - bsize
            else:
                self.buffer = self.buffer[consumed:]


    async def read_data(self):
        while True:
            data = await self.reader.read(self.read_size)
            if not data:
                # Disconnected
                self.running = False
                break

            self.buffer.extend(data)
            await self.new_data_arrived()


    async def ping(self):
        while self.running:
            await asyncio.sleep(30)

            header = Header(type=Type.TVT_HEAD.value, size=0)
            buf = create_header(header)
            assert len(buf) == SIZE_header

            self.writer.write(buf)
            await self.writer.drain()


    async def run(self):
        loop = asyncio.get_event_loop()

        try:
            reader, writer = await asyncio.open_connection(
                host=self.host,
                port=self.port,
                loop=loop)
        except ConnectionRefusedError:
            logging.warning(f"Can't connect to {self.host}:{self.port}")
            return
        except:
            logging.exception('Error happend')
            return

        self.handler = TvtHandler(
            writer=writer,
            username=self.username,
            password=self.password,
            channel=self.channel)

        self.reader = reader
        self.writer = writer

        await asyncio.gather(
            self.read_data(),
            self.ping())


def run(callback, host, port, channel, username=None, password=None):
    try:
        app = App(callback=callback, host=host, port=port, channel=channel, username=username, password=password)
        app.run()
    except:
        logging.exception('TVT run exception')
