import asyncio
import logging
import datetime
import json

from dataclasses import dataclass
from typing import Any
from pprint import pprint

from package.shared.constants import StreamSetting, GatewaySetting
from package.shared import models


@dataclass
class ESDataHandler():
    rxevents: Any = None

    async def handle_async(self, obj):
        now = datetime.datetime.now()
        session = models.SessionSingleton.instance().session

        # Users
        for user in obj['users']:
            await models.upsert_user(
                first_name=user['first_name'],
                last_name=user['last_name'],
                username=user['username'],
                password=user['password'])

        # Devices
        for pos in obj['poses']:
            await models.upsert_device(name=pos)

        # Gateways
        for gateway in obj['gateways']:
            gwobjs = {
                "cmd": "es-gateway-data",
                "obj": {
                    "gateways": [ gateway ],
                    "poses": obj['poses'],
                    "users": obj['users']
                }
            }

            if self.rxevents:
                self.rxevents.on_next(gwobjs)

            gw = await models.upsert_gateway(
                name=gateway['name'],
                serial=gateway['serial'],
                ldap_id=gateway['ldap_id'],
                path=gateway['path'],
                address=gateway['address'],
                port=int(gateway['port']),
                dataPort=int(gateway['dataPort']))

            await models.upsert_gateway_setting(gw, GatewaySetting.ES_GATEWAY_DATA, json.dumps(gwobjs))

            for stream in gateway['streams']:
                strm = await models.upsert_stream(
                    name=stream['name'],
                    serial=stream['serial'],
                    ldap_id=stream['ldap_id'],
                    path=stream['path'],
                    klass=stream['klass'],
                    gateway=gw,
                    input_url=stream['input_url'],
                    is_live=stream['is_live'],
                    is_record=stream['is_record'],
                    is_forward=stream['is_forward'])

                # Pos-Stream
                for p in stream['poses']:
                    try:
                        dev = session.query(models.Device).filter(models.Device.name==p).first()
                    except:
                        logging.exception("Exception")
                        session.rollback()
                        dev = None

                    if dev is None:
                        continue

                    await models.upsert_stream_device(strm, dev)
  
                # Settings
                for p in stream['settings']:
                    await models.upsert_stream_setting(
                        stream=strm, key=p,
                        value=stream['settings'][p])

                await models.delete_stream_devices_older_than(strm, now)
                await models.delete_stream_settings_older_than(strm, now)

            await models.delete_streams_older_than(gw, now)

        await models.delete_users_older_than(now)
        await models.delete_devices_older_than(now)
        await models.delete_gateways_older_than(now)

    def handle(self, obj):
        asyncio.create_task(self.handle_async(obj))

    def on_next(self, pkt):
        logging.info('Handling es data')
        self.handle(pkt['objs'])

    def on_error(self):
        pass

    def on_completed(self):
        pass