from dataclasses import dataclass
import logging

import zmq
from zmq.asyncio import Context as ZmqContext
from zmq.asyncio import Socket as ZmqSocket


@dataclass
class IPCListener():
    ipc_ctx = ZmqContext = None
    ipc_sock = ZmqSocket = None


    def __post_init__(self):
        pass


    def handle_ipc(self, client, buf):
        pass


    def setup_ipc(self, url, ctx, type):
        assert self.ipc_ctx is None
        assert self.ipc_sock is None

        self.ipc_ctx = ctx
        self.ipc_sock = ctx.socket(type)
        self.ipc_sock.setsockopt(zmq.LINGER, 1)
        self.ipc_sock.bind(url)

        logging.info('IPCListener is working on: {}'.format(url))


    def close_ipc(self):
        assert self.ipc_sock is not None

        logging.info('IPCListener is closed')
        self.ipc_sock.close()


    async def run_ipc(self):
        assert self.ipc_ctx is not None
        assert self.ipc_sock is not None

        logging.info('IPCListener is running')
        while True:
            try:
                client, buf = await self.ipc_sock.recv_multipart()
                self.handle_ipc(client, buf)
            except Exception as e:
                logging.warning('IPC Exception ({}): {}'.format(type(e), e))
                break


@dataclass
class IPCWorker():
    ipc_ctx: ZmqContext = None
    ipc_sock: ZmqSocket = None
    name: str = None
    url: str = None

    def __post_init__(self):
        pass


    def handle_ipc(self, buf):
        pass


    async def send_ipc(self, buf):
        assert self.ipc_sock is not None
        await self.ipc_sock.send(buf)


    def setup_ipc(self, url, ctx, type, name):
        assert self.ipc_ctx is None
        assert self.ipc_sock is None
        assert len(name)>0
        assert len(url)>0
        assert ctx is not None
        assert type is not None

        self.name = name
        self.url = url

        self.ipc_ctx = ctx
        self.ipc_sock = ctx.socket(type)
        self.ipc_sock.setsockopt(zmq.IDENTITY, name.encode())
        self.ipc_sock.setsockopt(zmq.LINGER, 1)
        self.ipc_sock.connect(url)

        logging.info('IPCWorker({}) is setup'.format(self.name))


    def close_ipc(self):
        assert self.ipc_sock is not None

        logging.info('IPCWorker({}) is closed'.format(self.name))
        self.ipc_sock.close()


    async def run_ipc(self):
        assert self.ipc_sock is not None

        logging.info('IPCWorker({}) is running'.format(self.name))
        while True:
            try:
                buf = await self.ipc_sock.recv()
                if not buf:
                    break

                self.handle_ipc(buf)
            except zmq.error.ZMQError:
                logging.exception(f'ZMQ socket error')
                break
            except Exception:
                logging.exception('IPCWorker exception')
                break
